package ap.proof.goal;

import ap.terfor.ConstantTerm
import ap.terfor.conjunctions.{Conjunction, Quantifier}
import ap.terfor.substitutions.VariableSubst
import ap.util.{Debug, PlainRange}
import ap.proof.tree.{ProofTree, ProofTreeFactory}

object QuantifierTask {
  
  private val AC = Debug.AC_COMPLEX_FORMULAS_TASK
  
}

abstract class QuantifierTask(_formula : Conjunction, _age : Int)
               extends FormulaTask(_formula, _age) {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(QuantifierTask.AC,
                   !formula.isTrue && !formula.isFalse &&
                   !formula.isLiteral && !formula.isNegatedConjunction &&
                   !formula.quans.isEmpty)
  //////////////////////////////////////////////////////////////////////////////
    
  /**
   * The name prefix to use for generated constants
   */
  protected def constantNameBase : String
  
  /**
   * Compute the new set of eliminated constants for a goal, given the set
   * of new constants (when carrying out this task)
   */
  protected def newEliminatedConstants(goal : Goal,
                                       newConstants : Iterable[ConstantTerm])
                                                            : Set[ConstantTerm]
    
                                                            
  protected def instantiateWithConstants(keepThisTask : Boolean,
                                         goal : Goal,
                                         ptf : ProofTreeFactory) : ProofTree = {
    val (quan, num) = firstQuans(formula)
    val constants = for (i <- PlainRange(num))
                    yield new ConstantTerm (constantNameBase + goal.age + "_" + i)

    val newOrder = (goal.order /: constants)((o, c) => o.extend(c, Set.empty))
    val newBindingContext = goal.bindingContext.addAndContract(constants, quan)
    val newConstantFreedom = if (quan == Quantifier.ALL)
                               goal.constantFreedom addTopStatus constants
                             else
                               goal.constantFreedom
    
    val newVocabulary = Vocabulary(newOrder, newBindingContext, newConstantFreedom)
    
    val newElimConstants = newEliminatedConstants(goal, constants)
    
    val instantiatedConjTask = {
      val instantiatedConj =
        new VariableSubst (0, constants, newOrder) (formula.unquantify(num))
      Goal.formulaTasks(instantiatedConj, goal.age,
                        newElimConstants, newVocabulary, goal.settings)
    }

    val thisTask : Iterable[PrioritisedTask] = if (keepThisTask)
                                                 goal.formulaTasks(formula)
                                               else
                                                 List()
    
    ptf.quantify(ptf.updateGoal(newElimConstants,
                                newVocabulary,
                                instantiatedConjTask ++ thisTask,
                                goal),
                 quan, Set.empty ++ constants, goal.vocabulary, newOrder)
  }
                                                              
  /**
   * Determine the leading quantifier of the conjunction and the number of the
   * same quantifiers in a row
   *
   * TODO: do this in a smarter way (determine which quantifiers can be permuted)?   
   */
  private def firstQuans(f : Conjunction) : (Quantifier, Int) = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(QuantifierTask.AC, !f.quans.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
                                                                
    val quans = f.quans
                                                                  
    val quan = quans.last
    val num = quans.size - quans.lastIndexOf(quan.dual) - 1
    (quan, num)
  }                                                            

  protected val name : String = "QuantifiedFor"

}
